import greenfoot.*;  // (World, Actor, GreenfootImage, and Greenfoot)
import java.util.List;

/**
 * A variation of an actor that maintains precise location (using doubles for the co-ordinates
 * instead of ints). It also maintains a current movement in form of a movement vector.
 * 
 * @author Poul Henriksen
 * @author Michael Kolling
 * 
 * @version 2.1
 * 
 * modified by Romuald Apothélos and Boris Méan.
 */
public abstract class SmoothMover extends Actor
{
    private Vector movement;
    private double exactX;
    private double exactY;
    
    public SmoothMover()
    {
        this(new Vector());
    }
    
    /**
     * Create new thing initialised with given speed.
     */
    public SmoothMover(Vector movement)
    {
        this.movement = movement;
    }
    
    /**
     * Move in the current movement direction.
     */
    public void move() 
    {
        exactX = exactX + movement.getX();
        exactY = exactY + movement.getY();
        super.setLocation((int) exactX, (int) exactY);
    }

    /**
     * Set the location using exact (double) co-ordinates.
     */
    public void setLocation(double x, double y) 
    {
        exactX = x;
        exactY = y;
        super.setLocation((int) x, (int) y);
    }
    
    /**
     * Set location. Redefinition of the standard Greenfoot method to make sure
     * the exact co-ordinates are updated in sync.
     */
    public void setLocation(int x, int y) 
    {
        exactX = x;
        exactY = y;
        super.setLocation(x, y);
    }

    /**
     * Return the exact x co-ordinate (as a double).
     */
    public double getExactX() 
    {
        return exactX;
    }

    /**
     * Return the exact y co-ordinate (as a double).
     */
    public double getExactY() 
    {
        return exactY;
    }

    /**
     * Modify the current movement by adding a new vector to the existing movement.
     */
    public void addForce(Vector force) 
    {
        movement.add(force);
    }
    
    /**
     * Accelerate the speed of this mover by the given factor. (Factors less than 1 will
     * decelerate.) The direction remains unchanged.
     */
    public void accelerate(double factor)
    {
        movement.scale(factor);
        if (movement.getLength() < 0.15) {
            movement.setNeutral();
        }
    }
    
    /**
     * Return the speed of this actor.
     */
    public double getSpeed()
    {
        return movement.getLength();
    }
    
    /**
     * Return the current movement of this object (as a vector).
     */
    public Vector getMovement() 
    {
        return movement;
    }
    /**
     * Gravity - apply the gravity to the Player
     */
    public void gravity(Vector movement,boolean ground)
    {
        
        if(ground && movement.getY()>-1){
            addForce (new Vector (0.0,-movement.getY()));
        } else
        {
            addForce (new Vector (0.0, 0.1));
        }
    }
    /**
     * OnGround - check if the player touche the ground or a platform
     */
    public boolean onGround(Vector movement, int height)
    {
       List<Platform> platfrom = (List<Platform>) getWorld().getObjects(Platform.class);
       List<Player> player = (List<Player>) getWorld().getObjects(Player.class);
       List<Obstacle> obstacle = (List<Obstacle>) getWorld().getObjects(Obstacle.class);

       
        for (Platform plat : platfrom) 
        {
            if (plat.getX()+50>this.getX() && plat.getX()-50<this.getX() && plat.getY()-4<=this.getY()+height && plat.getY()+4>=this.getY()+height )
            {
               if(movement.getY()>0){setLocation(this.getX(),plat.getY()-height);}
               return true;
            }
        }
        for (Player play : player) 
        {
            if (play.getX()+15>this.getX() && play.getX()-15<this.getX() && play.getY()-40<=this.getY()+height && play.getY()+40>=this.getY()+height && play != this)
            {
               return true;
            }
        }
        for (Obstacle obj : obstacle) 
        {
            if (obj.getX()+obj.getHeight()+10>this.getX() && obj.getX()-obj.getHeight()-10<this.getX() && obj.getY()-obj.getHeight()<=this.getY()+height && obj.getY()+obj.getHeight()>=this.getY()+height && obj != this)
            {
               if(movement.getY()>0){setLocation(this.getX(),obj.getY()-(obj.getHeight()+height));}
               return true;
            }
        }
        if(getY()>=(600-height)){
            setLocation(this.getX(),600-height);
            return true;
        }
        return false;
    }
    /**
     * OnWallL - check if the player is near a wall
     */
    public int onWall(Vector movement, int width )
    {
        List<Wall> wall = (List<Wall>) getWorld().getObjects(Wall.class);
        List<Player> player = (List<Player>) getWorld().getObjects(Player.class);
        List<Obstacle> obstacle = (List<Obstacle>) getWorld().getObjects(Obstacle.class);
        
        for (Wall wal : wall) 
        {
            if (wal.getX()-20<this.getX()+width  && wal.getX()>this.getX()+width && wal.getY()-90<this.getY() && wal.getY()+90>this.getY())
            {
               return 2;
            }
            if (wal.getX()+20>this.getX()-width  && wal.getX()<this.getX()-width && wal.getY()-90<this.getY() && wal.getY()+90>this.getY())
            {
               return 1;
            }
        }
        for (Player play : player) 
        {
            if (play.getX()>this.getX()+width && play.getX()-16<this.getX()+width && play.getY()-80<=this.getY() && play.getY()+80>=this.getY() && play != this)
            {
               return 2;
            }
            if (play.getX()+16>this.getX()-width && play.getX()<this.getX()-width && play.getY()-80<=this.getY() && play.getY()+80>=this.getY() && play != this)
            {
               return 1;
            }
        }
        for (Obstacle obj : obstacle) 
        {
            if (obj.getX()>this.getX()+width && obj.getX()-obj.geWidth()-1<this.getX()+width && obj.getY()-obj.geWidth()<=this.getY() && obj.getY()+obj.geWidth()>=this.getY() && obj != this)
            {
               if(movement.getX()>0){setLocation(obj.getX()-(obj.geWidth()+width),this.getY());}
               return 2;
            }
            if (obj.getX()+obj.geWidth()+1>this.getX()-width && obj.getX()<this.getX()-width && obj.getY()-obj.geWidth()<=this.getY() && obj.getY()+obj.geWidth()>=this.getY() && obj != this)
            {
               if(movement.getX()<0){setLocation(obj.getX()+(obj.geWidth()+width),this.getY());}
               return 1;
            }
        }
        if(this.getX()<=20){return 1;};
        if(this.getX()>=1180){return 2;};
        return 0;
    }
    public void onTop(Vector movement, int height )
    {
        List<Platform> platfrom = (List<Platform>) getWorld().getObjects(Platform.class);
        List<Player> player = (List<Player>) getWorld().getObjects(Player.class);
        List<Obstacle> obstacle = (List<Obstacle>) getWorld().getObjects(Obstacle.class);
        
        for (Platform plat : platfrom) 
        {
            if (plat.getX()+50>this.getX() && plat.getX()-50<this.getX() && plat.getY()-4<=this.getY()-height && plat.getY()+4>=this.getY()-height )
            {
               if(plat.clear()==false)
               {
                   if(movement.getY()<0){addForce (new Vector (0.0, -1.2*movement.getY()));}
               }
            }
        }
        for (Player play : player) 
        {
            if (play.getX()+15>this.getX() && play.getX()-15<this.getX() && play.getY()-40<=this.getY()-height && play.getY()+40>=this.getY()-height && play != this)
            {
               if(movement.getY()<0){addForce (new Vector (0.0, -1.2*movement.getY()));}
            }
        }
        for (Obstacle obj : obstacle) 
        {
            if (obj.getX()+obj.getHeight()+10>this.getX() && obj.getX()-obj.getHeight()-10<this.getX() && obj.getY()-obj.getHeight()<=this.getY()-height && obj.getY()+obj.getHeight()>=this.getY()-height && obj != this)
            {
               if(movement.getY()<0){addForce (new Vector (0.0, -1.2*movement.getY()));}
            }
        }
    }
    /**
     * explod - affect all the object in a radius
     */
    public void explod(Actor killer)
    {
        getWorld().addObject(new Explosion(), getX(), getY()); 
        List<Player> player = getObjectsInRange(150, Player.class);
        List<Obstacle> obstacle = getObjectsInRange(150, Obstacle.class);
        for (Player play : player) 
        {
            double dx = play.getExactX() - this.getExactX();
            double dy = play.getExactY() - this.getExactY();
            Vector force = new Vector (dx, dy);
            double distance = Math.sqrt (dx*dx + dy*dy);
            double acceleration = 2000 / (distance * distance/10);
            double damage = 1000 / (distance * distance/100);
            force.setLength (acceleration);
            play.addForce (force);
            play.hurt((int) damage,killer);
        }
        for (Obstacle obs : obstacle) 
        {
            double dx = obs.getExactX() - this.getExactX();
            double dy = obs.getExactY() - this.getExactY();
            Vector force = new Vector (dx, dy);
            double distance = Math.sqrt (dx*dx + dy*dy);
            double acceleration = 2000 / (distance * distance/10);
            double damage = 1000 / (distance * distance/100);
            force.setLength (acceleration);
            obs.addForce (force);
            obs.destroy((int) damage, killer);
        }
    }
    /**
     * Setfire - put fire around the object
     */
    public void setfire(Actor iniciateur)
    {
        for(int i = 0;i<5;i++)
        {
            getWorld().addObject(new GroundSticker(0,iniciateur),getX()-40+i*20,getY());
        }
    }
}
